from typing import Tuple, List

import numpy as np
import mysql

from ...logger import logging
from .simulation_db_types import TABLE_NAMES
from ...config import MYSQL_USER, MYSQL_PW, MYSQL_HOST, MYSQL_PORT
from acss_core.utils.SqlDBConnector import SqlDBConnector

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)

if MYSQL_USER == None:
    raise Exception(f"Environment Value SIM_SQL_USER isn't set. Please set the value in .env file.")

if MYSQL_PW == None:
    raise Exception(f"Environment Value SIM_SQL_PW isn't set. Please set the value in .env file.")

if MYSQL_HOST == None:
    raise Exception(f"Environment Value SIM_SQL_HOST isn't set. Please set the value in .env file.")

_logger.debug(f"Environment SIM_SQL_HOST: {MYSQL_HOST}")


class PetraAdapter():
    def __init__(self, custom_db_connector):

        self._excluded_prefix_cor_names = []
        self.db_connector = custom_db_connector if custom_db_connector is not None else SqlDBConnector(user=MYSQL_USER, pw=MYSQL_PW, host=MYSQL_HOST, port=MYSQL_PORT, database='sim_db')
        self.db_connector.check_if_sql_db_table_exists(TABLE_NAMES.MACHINE_PARMS)
        self.active_package_id = None
        self.service_type = 'petra_sim_adapter'

    def set_cor_name_filter(self, prefix_names: List[str]):

        self._excluded_prefix_cor_names = prefix_names

    def cor_names_filter(self, names):
        if not self._excluded_prefix_cor_names:
            return names
        new_names = []
        for name in names:
            is_excluded = False
            for prefix in self._excluded_prefix_cor_names:
                if name.startswith(prefix):
                    is_excluded = True
                    break
            if not is_excluded:
                new_names.append(name)

        return new_names

    def set_machine_params(self, names, values):
        try:
            data = [(value, name) for name, value in zip(names, values)]
            cursor = self.db_connector.sql_con.cursor()
            cursor.executemany("Update machine_params set value = %s where param = %s", data)
            cursor.close()
        except mysql.connector.Error as error:
            _logger.info(f"Failed to update record to database rollback: {error}")
            self.db_connector.sql_con.rollback()

    def get_machine_params(self):
        sql_select_Query = f"select param, value from {TABLE_NAMES.MACHINE_PARMS}"
        cursor = self.db_connector.sql_con.cursor()
        cursor.execute(sql_select_Query)
        res = {res[0]: res[1] for res in cursor.fetchall()}
        self.db_connector.sql_con.commit()
        cursor.close()
        return res

    def get_hcor_device_names(self) -> List[str]:
        """[summary]
        Returns all getable an settable horizontal corrector names
        :return: List of horizontal corrector device names in uppercase
        :rtype: List[str]
        """
        raise NotImplemented()

    def get_vcor_device_names(self) -> List[str]:
        """[summary]
        Returns all getable an settable vertical corrector names
        :return: List of horizontal vertical device names in uppercase
        :rtype: List[str]
        """
        raise NotImplemented()

    def get_hcors(self, names: List[str], is_group_call=True) -> List[float]:
        """[summary]
        Gets the strengths of horizontal corrector names,
        :param names: Names of the horizontal correctors in uppercase.
        :type names: List[str]
        :param is_group_call: Read all correctors parallel which is faster if True else read serial, defaults to True
        :type is_group_call: bool, optional
        :return: List of strengths sorted by names
        :rtype: List[float]
        """
        raise NotImplemented()

    def get_vcors(self, names: List[str], is_group_call=True) -> List[float]:
        """[summary]
        Gets the strengths of vertical corrector names,
        :param names: Names of the vertical correctors. Have to be uppercase.
        :type names: List[str]
        :param is_group_call: Reads all correctors parallel which is faster if True else reads serial, defaults to True
        :type is_group_call: bool, optional
        :return: List of strengths sorted by names
        :rtype: List[float]
        """
        raise NotImplemented()

    def set_vcors(self, names: List[str], strengths: List[float], current_offsets: List[float] = None, is_group_call=True):
        """[summary]
        Set the vertical correctors.
        :param names: Names of the vertical correctors to be set in uppercase.
        :type names: List[str]
        :param in_strengths: List of strenghts to be set.
        :type in_strengths: List[float]
        :param current_offsets: List of offsets that will be added to the bpm's current, default to None
        :type current_offsets: List[float]
        :param is_group_call: Writes all correctors parallel which is faster if True else writes serial, defaults to True
        :type is_group_call: bool, optional
        """
        raise NotImplemented()

    def set_hcors(self, names: List[str], strengths: List[float], current_offsets: List[float] = None, is_group_call=True):
        """[summary]
        Set the horizontal correctors.
        :param names: Names of the horizontal correctors to be set in uppercase.
        :type names: List[str]
        :param in_strengths: List of strenghts to be set.
        :type in_strengths: List[float]
        :param current_offsets: List of offsets that will be added to the bpm's current, default to None
        :type current_offsets: List[float]
        :param is_group_call: Writes all correctors parallel which is faster if True else writes serial, defaults to True
        :type is_group_call: bool, optional
        """
        raise NotImplemented()

    def get_bpms(self) -> Tuple[np.ndarray, np.ndarray, List[str]]:
        """[summary]
        Gets the bpm coordinates and names.
        :return: A Tuple (x, y, names) where x, y are the coordinates of the bpm.
        """
        raise NotImplemented()

    def get_ddxy(self, n_turns=10) -> Tuple[np.ndarray, List[str]]:
        raise NotImplemented()

    def get_dd_sum(self, n_turns=10) -> Tuple[np.ndarray, List[str]]:
        raise NotImplemented()



import os
import argparse


def set_service_by_env_values():
    name = os.environ.get('SERVICE_NAME')
    mode = os.environ.get('SERVICE_MODE')
    if name == None:
        raise Exception('Environment SERVICE_NAME is not set.')
    read = None
    if mode == 'sim':
        from .TineSimAdapter import TineSimReader
        read = TineSimReader.create_for_petra()
    elif mode == 'prod':
        from .TineAdapter import TineReader
        read = TineReader()
    else:
        raise Exception('Environment SERVICE_MODE have to set to "sim" or "prod".')
    return name, read, mode


def set_adapter_by_arg_parser():
    parser = argparse.ArgumentParser(description='Short sample app')
    parser.add_argument('--mode', help="mode can be 'prod' or 'sim'")
    options = parser.parse_args()
    read = None

    if options.mode == 'sim':
        from .TineSimAdapter import TineSimReader
        read = TineSimReader.create_for_petra()
    elif options.mode == 'prod':
        from .TineAdapter import TineReader
        read = TineReader()
    else:
        raise Exception("Argument --mode isn't set. Possible values are 'prod' for production or 'sim' for simulation")
    return read, options.mode



import numpy as np
import time
import csv
from typing import Dict, Tuple

from .bpm_fit import correct_bpm
from ...logger import logging


_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)


class BPM:
    def __init__(self, bpm_type, path_to_calibr_files):
        bpm_file = bpm_type+'.par'
        calibr_file = path_to_calibr_files / "common" / bpm_file
        colmat = np.loadtxt(calibr_file)
        self.cx = [r[0] for r in colmat]
        self.cy = [r[1] for r in colmat]

    def correct_x(self, x):
        x_ = 0.0
        for i in range(len(self.cx)):
            x_ += self.cx[i] * x**i
        return x_

    def correct_y(self, y):
        y_ = 0.0
        for i in range(len(self.cy)):
            y_ += self.cy[i] * y**i
        return y_


class BPMAdapter:
    def __init__(self, read, path_to_calibr_files, path_to_constants_file):
        self.path_to_calibr_files = path_to_calibr_files
        self.read = read

        self.bpm_names = self.get_bpm_names()
        self.bpm_names = [p.upper() for p in self.bpm_names]

        # Constants to calculate the bpm beam current
        self.constants: Dict[str, Tuple[int, int]] = {}
        if path_to_constants_file:
            with open(path_to_constants_file) as csvfile:
                reader = csv.DictReader(csvfile)
                for row in reader:
                    self.constants[row['name']] = (row['c0'], row['c1'])
            _logger.debug(f"Constants loaded: length {len(self.constants)}")

        with open(self.path_to_calibr_files / "BPMsettings.csv") as csvfile:
            reader = csv.DictReader(csvfile)
            devices = []
            bpm_types = []
            for row in reader:
                devices.append(row['Device'])
                bpm_types.append(row['BPMTYP'])

        self.bpm_info = {device: bpm_type for device, bpm_type in zip(devices, bpm_types)}

        self.load_calibration(set(bpm_types))

    def get_bpm_names(self):
        return self.read("/PETRA/LBRENV", "BPM_SWR_13", "DEVICES")

    def load_calibration(self, bpm_types):
        self.PQ = {}

        for bpm_t in bpm_types:
            bpm = BPM(bpm_t, path_to_calibr_files=self.path_to_calibr_files)
            self.PQ[bpm_t] = (bpm.cx, bpm.cy)

        return

    def get_names(self, start_with):
        start_i = self.bpm_names.index(start_with)
        names = self.bpm_names[start_i:]+self.bpm_names[:start_i]  # realign names to start from start_with
        return start_i, names

    def _read(self, n_turns=10, start_with="BPM_SOR_61", is_corrected=True):
        start_i, names = self.get_names(start_with)

        res = np.zeros((len(names), n_turns, 2))

        for i, name in enumerate(names):
            for k in range(3):
                try:
                    res[i, :, 0] = np.array(self.read('/PETRA/LBRENV', name, 'DD_X', size=n_turns, mode='SYNC'))
                    res[i, :, 1] = np.array(self.read('/PETRA/LBRENV', name, 'DD_Y', size=n_turns, mode='SYNC'))
                    break
                except Exception as e:
                    print(f'attempt {k}:', e)
                    time.sleep(1)

        if is_corrected:
            res = self.correct(res, names)
        return res, names

    def read_sum(self, n_turns=10, start_with="BPM_SOR_61", is_corrected=True):
        start_i, names = self.get_names(start_with)

        res = np.zeros((len(names), n_turns, 1))

        for i, name in enumerate(names):
            for k in range(3):
                try:
                    res[i, :, 0] = np.array(self.read('/PETRA/LBRENV', name, 'DD_SUM', size=n_turns, mode='SYNC'))
                    break
                except Exception as e:
                    print(f'attempt {k}:', e)
                    time.sleep(1)

        return res, names

    def correct(self, res, names, scale=1e-6):
        res *= scale
        print(res.shape)
        res_cor = np.empty_like(res)
        for i, name in enumerate(names):
            PQ = self.PQ[self.bpm_info[name]]
            x, y = correct_bpm(res[i, :, 0], res[i, :, 1], PQ[0], PQ[1])
            res_cor[i, :, 0] = x
            res_cor[i, :, 1] = y
        return res_cor/scale

    def get_offsets(self, start_with="BPM_SOR_61"):
        name0 = self.bpm_names[0]
        x = np.array(self.read('/PETRA/REFORBIT', name0, 'CORR_X_BBA'))
        y = np.array(self.read('/PETRA/REFORBIT', name0, 'CORR_Y_BBA'))
        res = np.vstack((x, y)).T
        res = res[:-2]

        start_i, names = self.get_names(start_with)
        return np.roll(res, -start_i, axis=0), names

    def get_offsets_go(self, start_with="BPM_SOR_61"):
        name0 = self.bpm_names[0]
        x = np.array(self.read('/PETRA/REFORBIT', name0, 'CORR_X_BBAGO'))
        y = np.array(self.read('/PETRA/REFORBIT', name0, 'CORR_Y_BBAGO'))
        res = np.vstack((x, y)).T
        res = res[:-2]
        start_i, names = self.get_names(start_with)
        return np.roll(res, -start_i, axis=0), names

    def get_orbit(self, start_with="BPM_SOR_61"):
        """ In meter

        :param start_with: _description_, defaults to "BPM_SOR_61"
        :type start_with: str, optional
        :return: _description_
        :rtype: _type_
        """
        name0 = self.bpm_names[0]
        x = np.array(self.read('/PETRA/REFORBIT', name0, 'SA_X'))
        y = np.array(self.read('/PETRA/REFORBIT', name0, 'SA_Y'))
        res = np.vstack((x, y)).T
        res = res[:-2]
        start_i, names = self.get_names(start_with)
        return np.roll(res, -start_i, axis=0), names

    def get_bpm_beam_currents(self, n_turns=10):
        if self.c0 == None or self.c1 == None:
            _logger.error(f"Constants c0, c1 are not set. c0={self.c0} and c1={self.c1}")
            return None, None
        sum_signal, names = self.read_sum(n_turns=n_turns)
        n_bpms = len(names)
        currents = np.zeros([n_bpms, n_turns])

        for i, name in enumerate(names):
            c0, c1 = self.constants.get(name)
            currents[i] = c0 + c1 * sum_signal[i, :, 0]
        return currents, names


if __name__ == "__main__":
    adapter = BPMAdapter()
    data, names = adapter.get_orbit()



import numpy as np
import PyTine as pt
from K2I2K_os import K2I2K_os
from typing import List


class TineAdapter():
    def __init__(self, energy=6.063) -> None:
        self.energy = energy

    def find_latest_server(prefix, servers):
        latest_server = None
        latest_version = 0
        for server in servers:
            if server.startswith(prefix):
                version = int(server[len(prefix):])
                if version > latest_version:
                    latest_version = version
                    latest_server = server
        return latest_server

    def find_latest_optic_server():
        servers = pt.list('Common')['servers']
        return TineAdapter.find_latest_server('K2I2KL', servers), TineAdapter.find_latest_server('OPTICL', servers)

    def strength2current(self, names: List, strengths: List) -> np.ndarray:
        kk = K2I2K_os(names, psStrength=strengths, energy=self.energy, debug=None)
        return kk.psCurrent

    def current2strength(self, names: List, currents: List) -> np.ndarray:
        kk = K2I2K_os(names, psCurrent=currents, energy=self.energy, debug=None)
        return kk.psStrength


class TineWriter(TineAdapter):
    def __init__(self, energy=6.063):
        super().__init__(energy=6.063)

    def __call__(self, channel: str, device: str, _property, **kwargs):
        addr = f"{channel}/{device}"

        if _property == "Strength.Soll":
            strengths = kwargs.get("input")
            if strengths is None:
                raise ValueError("input is not set.")

            filtered_kwargs = {key: value for key, value in kwargs.items() if key != 'input'}
            if channel == "/PETRA/Cms.PsGroup":
                size = pt.get(addr, "GroupSize")['data']
                names = pt.get(addr, "GroupDevices", size=size)['data']
                currents = self.strength2current(names, strengths)
                print("New Currents:")
                print(*list(zip(names,currents)), sep='\n')                


                pt.set(addr, "Strom.Soll", input=list(currents), **filtered_kwargs)
                return
            elif channel == "/PETRA/Cms.MagnetPs":
                current = self.strength2current([device], [strengths])[0]
                pt.set(addr, "Strom.Soll", input=current, **filtered_kwargs)
                return
            else:
                raise NotImplemented()

        pt.set(addr, _property, **kwargs)

    def commit():
        pass


class TineReader(TineAdapter):
    def __init__(self, energy=6.063):
        super().__init__(energy=6.063)

    def __call__(self, channel: str, device: str, _property: str, **kwargs):
        addr = f"{channel}/{device}"

        if _property == "Strength.Soll":
            if channel == "/PETRA/Cms.PsGroup":
                #size = kwargs.get('size', pt.get(addr, "GroupSize")['data'])
                size = pt.get(addr, "GroupSize")['data']

                names = pt.get(addr, "GroupDevices", size=size)['data']
                currents = pt.get(addr, "Strom.Soll", size=size)['data']
                return self.current2strength(names, currents)
            elif channel == "/PETRA/Cms.MagnetPs":
                current = pt.get(addr, 'Strom.Soll', **kwargs)['data']
                return self.current2strength([device], [current])[0]

        data = pt.get(addr, _property, **kwargs)['data']

        if channel == "/PETRA/REFORBIT":
            if _property in {"SA_X", "SA_Y", 'CORR_X_BBA', 'CORR_Y_BBA', 'CORR_X_BBAGO', 'CORR_Y_BBAGO'}:
                data = data * 10e-9 # nm to m
        return data
